/*----------------------------------------------------------------------
  File    : gramod2.c
  Contents: graphical model management (based on directed graphs)
            parameter estimation and local learning functions
  Author  : Christian Borgelt
  History : 1995.11.03 file created as ines.c
            2002.03.03 network induction functions moved out of ines.c
            2002.03.30 function gm_evalx added (from sian.c)
            2002.07.04 local structure learning (re)added
            2006.03.04 these functions made a separate file
            2007.01.29 bug in function gm_params fixed (gm_setnorm)
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#ifndef GM_EVAL
#define GM_EVAL
#endif
#ifndef GM_INDUCE
#define GM_INDUCE
#endif
#include "gramod.h"

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *snames[] = { /* search method names */
  /* GM_NONE     0 */ "no search, only parameter estimation",
  /* GM_OWST     1 */ "optimum weight spanning tree construction",
  /* GM_EXTST    2 */ "optimum weight spanning tree extension",
  /* GM_TOPORD   3 */ "greedy condition selection on topological order",
  /* GM_NOLOOP   4 */ "greedy condition selection avoiding loops",
  /* GM_UNRES    5 */ "greedy condition selection without restrictions",
  /* GM_CITEST   6 */ "conditional independence tests",
  /* GM_SIAN     7 */ "hypertree simulated annealing",
  /* GM_UNKNOWN  8 */ "unknown search method"
};

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

int gm_estim (GRAMOD *gm, TABLE *table, int distuv)
{                               /* --- estimate model parameters */
  ATTID i, k;                   /* loop variables */
  GMATT *att;                   /* to traverse the attributes */
  TUPLE *tpl;                   /* to traverse the tuples */
  float wgt;                    /* weight of the instantiation */

  assert(gm && table);          /* check the function arguments */
  for (i = tab_tplcnt(table); --i >= 0; ) {
    tpl = tab_tpl(table, i);    /* traverse the tuples */
    tpl_toas(tpl);              /* copy tuple to attribute set */
    wgt = tpl_getwgt(tpl);      /* get the instantiation weight */
    for (att = gm->atts +(k = gm->attcnt); --k >= 0; )
      if ((--att)->ptree && (pt_aggr(att->ptree, wgt, distuv) < 0))
        return -1;              /* aggregate the instantiation weight */
  }                             /* for all attributes */
  return 0;                     /* return 'ok' */
}  /* gm_estim() */

/*--------------------------------------------------------------------*/

double gm_eval (GRAMOD *gm, int measure, double *params)
{                               /* --- evaluate a graphical model */
  ATTID i;                      /* loop variable */
  GMATT *att;                   /* to traverse the attributes */

  assert(gm);                   /* check the function arguments */
  gm->eval = 0;                 /* init. the evaluation result */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
    if (!(--att)->ptree) continue;
    att->eval = pt_eval(att->ptree, measure, params);
    if (isnan(att->eval)) return NAN;
    gm->eval += att->eval;      /* evaluate all prob./poss. trees */
  }
  return gm->eval;              /* return sum of evaluation results */
}  /* gm_eval() */

/*--------------------------------------------------------------------*/

double gm_evalx (GRAMOD *gm, TABLE *table, double res[])
{                               /* --- evaluate a graphical model */
  ATTID  i;                     /* loop variable */
  double tsr[3];                /* tuple specific results */
  double imposs;                /* number of impossible tuples */
  TUPLE  *tpl;                  /* to traverse the tuples */
  float  wgt;                   /* instantiation weight */

  assert(gm && table);          /* check the function arguments */
  if (gm_check(gm, -1, 0) != 0) /* if there are loops, */
    return -1;                  /* abort the function */
  res[0] = res[1] = res[2] = imposs = 0;
  for (i = tab_tplcnt(table); --i >= 0; ) {
    tpl = tab_tpl(table, i);    /* traverse the tuples */
    tpl_toas(tpl);              /* copy tuple to attribute set */
    gm_execx(gm, tsr);          /* execute the graphical model */
    wgt = tpl_getwgt(tpl);      /* get the instantiation weight */
    if (gm->type & GM_POSS) {   /* aggregate possibilities */
      res[0] += wgt *tsr[0];    /* (simply sum the weighted */
      res[1] += wgt *tsr[1];    /* evaluation results) */
      res[2] += wgt *tsr[2]; }
    else {                      /* aggregate probabilities */
      if (tsr[0] > 0) res[0] += wgt *log(tsr[0]);
      if (tsr[1] > 0) res[1] += wgt *log(tsr[1]);
      if (tsr[2] > 0) res[2] += wgt *log(tsr[2]);
    }                           /* (compute log-likelihood) */
    if (tsr[2] <= 0)            /* if the maximum prob./poss. is 0, */
      imposs += wgt;            /* count the tuple as impossible */
  }
  return imposs;                /* return the number of */
}  /* gm_evalx() */             /* impossible tuples */

/*--------------------------------------------------------------------*/

double gm_local (GRAMOD *gm, int lwise, double minimp)
{                               /* --- do local structure learning */
  ATTID i;                      /* loop variable */
  GMATT *att;                   /* to traverse the attributes */

  assert(gm);                   /* check the function arguments */
  gm->eval = 0;                 /* init. the evaluation result */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
    if (!(--att)->ptree) continue;
    att->eval = pt_local(att->ptree, lwise, minimp);
    if (isnan(att->eval)) return NAN;
    gm->eval += att->eval;      /* evaluate all prob./poss. trees */
  }
  return gm->eval;              /* return sum of evaluation results */
}  /* gm_local() */

/*--------------------------------------------------------------------*/

const char* gm_sname (int search)
{                               /* --- get name of search method */
  if ((search < GM_NONE) || (search > GM_UNKNOWN))
    search = GM_UNKNOWN;        /* check the search method index */
  return snames[search];        /* return the search method name */
}  /* gm_sname() */

/*--------------------------------------------------------------------*/

double gm_params (GRAMOD *gm, TABLE *table, int measure, double *params)
{                               /* --- induce by learning parameters */
  double tplcnt;                /* number of tuples (weight) */

  assert(gm && table);          /* check the function arguments */
  if (gm->attcnt <= 0)          /* if there are no attributes, */
    return 0.0;                 /* simply abort the function */
  tplcnt = tab_getwgt(table, 0, TPLID_MAX);
  if ((gm->type & GM_POSS)      /* for poss. graphical models */
  &&  (tab_opc(table, TAB_COND) != 0))
    return NAN;                 /* compute one point coverages */
  gm_clear(gm, 1);              /* clear the graphical model params. */
  gm_setnorm(gm, tplcnt);       /* set number of tuples for norm. */
  if ((gm_estim(gm, table, 0) != 0)
  ||  isnan(gm_eval(gm, measure, params)))
    return NAN;                 /* estimate parameters and evaluate */
  return gm->eval;              /* return the evaluation result */
}  /* gm_params() */
