/*----------------------------------------------------------------------
  File    : neval.c
  contents: network evaluation
  Author  : Christian Borgelt
  History : 1997.03.02 file created
            1997.09.08 minor improvements
            1998.02.02 option -u (unknown values characters) added
            1998.06.23 adapted to modified attset functions
            1999.04.18 simplified using the new module 'io'
            1999.04.30 one point coverage computation added
            1999.12.15 option -c (compare to reference network) added
            2001.11.25 adapted to modified scanner functions
            2002.01.22 adapted to modified gramod and ptree modules
            2002.02.24 check for multiple assignment of stdin improved
            2002.04.04 adapted to new functions gm_consum and gm_parsum
            2003.01.16 optional approximation for poss. maximum removed
            2003.08.16 slight changes in error message output
            2006.03.12 detection and counting of reversed edges added
            2007.01.29 loop check removed for possibilistic models
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <math.h>
#include <assert.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef GM_PARSE
#define GM_PARSE
#endif
#include "gramod.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "neval"
#define DESCRIPTION "probabilistic/possibilistic network evaluation"
#define VERSION     "version 3.1 (2014.10.24)         " \
                    "(c) 1997-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse error */
#define E_LOOP      (-10)       /* condition loop */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_LOOP    -10 */  "condition loop detected",
  /*           -11 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for graphical model) */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static GRAMOD  *gramod = NULL;  /* graphical model */
static ATTSET  *refas  = NULL;  /* attribute set of reference network */
static GRAMOD  *refgm  = NULL;  /* reference graphical model */
static TABLE   *table  = NULL;  /* table to evaluate model on */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (table)  tab_delete(table, 0); \
  if (refgm)  gm_delete(refgm,  0); \
  if (refas)  as_delete(refas);     \
  if (gramod) gm_delete(gramod, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_net  = NULL;      /* name of network file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_ref  = NULL;      /* name of reference network file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  int     mode     = AS_ATT|AS_NOEXT; /* table file read mode */
  int     type     = GM_AUTO;   /* type of graphical model */
  int     diffsum  = 0;         /* compute sum of poss. degree diffs. */
  double  imposs   = 0;         /* number of impossible tuples */
  WEIGHT  lcorr    = 0;         /* Laplace correction */
  double  res[3];               /* aggregation of evaluation results */
  double  rgm[3], rtab[3];      /* tuple specific evaluation results */
  TUPLE   *tpl;                 /* to traverse the tuples */
  WEIGHT  wgt;                  /* instantiation weight */
  ATTID   addcnt, miscnt;       /* additional and missing conditions */
  ATTID   revcnt;               /* reversed conditions/edges */
  ATTID   m, c, x, y, z;        /* number of attributes/conditions */
  TPLID   n, r;                 /* number of data tuples */
  double  w, total;             /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] netfile", argv[0]);
    printf(         " [-d|-h hdrfile] tabfile [outfile]\n");
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c <fn>  compare to reference network in file <fn>\n");
    printf("-s       compute sum of possibility differences\n");
    printf("-L#      Laplace correction                     "
                    "(default: %g)\n", lcorr);
    printf("-n       number of tuple occurrences in last field\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("netfile  file containing network structure to evaluate\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write evaluation results to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': optarg  = &fn_ref;               break;
          case 's': diffsum = 1;                     break;
          case 'L': lcorr   = (WEIGHT)strtod(s, &s); break;
          case 'r': optarg  = &recseps;              break;
          case 'f': optarg  = &fldseps;              break;
          case 'b': optarg  = &blanks;               break;
          case 'C': optarg  = &comment;              break;
          case 'u': optarg  = &nullchs;              break;
          case 'n': mode   |= AS_WEIGHT;             break;
          case 'd': mode   |= AS_DFLT;               break;
          case 'h': optarg  = &fn_hdr;               break;
          default : error(E_OPTION, *(--s));         break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_net = argv[i]; break;
        case  1: fn_tab = argv[i]; break;
        case  2: fn_out = argv[i]; break;
        default: error(E_ARGCNT);  break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the program arguments */
  if ((k < 2) || (k > 3)) error(E_ARGCNT);
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  if (fn_ref && (strcmp(fn_ref, "-") == 0))
    fn_ref = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  ( fn_ref && !*fn_ref) i++;
  if  (!fn_net || !*fn_net) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_net) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan));

  /* --- parse graphical model --- */
  gramod = gm_parse(attset, scan, GM_AUTO|GM_ALL);
  if (!gramod || !scn_eof(scan, 1)) /* parse or create */
    error(E_PARSE, scn_name(scan)); /* a graphical model */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  type = gm_type(gramod);       /* get the aggregation mode */
  if ((type == GM_PROB) && (gm_check(gramod, -1, 0) < 0))
    error(E_LOOP);              /* check for a condition loop */
  gm_setlc(gramod, lcorr);      /* set the Laplace correction */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  table = tab_create("table", attset, tpl_delete);
  if (!table) error(E_NOMEM);   /* create a data table */
  k = tab_read(table, tread, mode);
  if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(table);        /* get the number of attributes */
  n = tab_tplcnt(table);        /* and the number of data tuples */
  w = tab_tplwgt(table);        /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- evaluate graphical model --- */
  res[0] = res[1] = res[2] = 0; /* init. the evaluation results */
  rgm[2] = 1; total = w;        /* and set the default maximum */
  t = clock();                  /* start the timer */
  if (type != GM_POSS) {        /* if probabilistic graphical model */
    fprintf(stderr, "evaluating tuples ... ");
    for (r = 0; r < n; r++) {   /* traverse the tuples */
      tpl = tab_tpl(table, r);  /* get the next tuple */
      wgt = tpl_getwgt(tpl);    /* and its weight and */
      tpl_toas(tpl);            /* copy it to the attribute set */
      gm_execx(gramod, rgm);    /* execute the graphical model */
      if (rgm[0] > 0) res[0] += wgt *log(rgm[0]);
      if (rgm[1] > 0) res[1] += wgt *log(rgm[1]);
      if (rgm[2] > 0) res[2] += wgt *log(rgm[2]);
      else imposs += wgt;       /* if the maximum probability is 0, */
    } }                         /* count the tuple as impossible */
  else if (!diffsum) {          /* if not to compute differences */
    fprintf(stderr, "evaluating tuples ... ");
    for (r = 0; r < n; r++) {   /* traverse the tuples */
      tpl = tab_tpl(table, r);  /* get the next tuple */
      wgt = tpl_getwgt(tpl);    /* and its weight and */
      tpl_toas(tpl);            /* copy it to the attribute set */
      gm_execx(gramod, rgm);    /* execute the graphical model */
      res[0] += wgt *rgm[0];    /* and sum the weighted */
      res[1] += wgt *rgm[1];    /* evaluation results */
      res[2] += wgt *rgm[2];    /* if the maximum possibility is 0, */
      if (res[2] <= 0) imposs += wgt;
    } }                         /* count the tuple as impossible */
  else {                        /* if to compute differences */
    fprintf(stderr, "computing one point coverages ... ");
    for (r = n = tab_reduce(table); --r >= 0; ) {
      tpl = tab_tpl(table, r);  /* traverse the reduced table */
      tpl_setxwgt(tpl, tpl_getwgt(tpl));
    }                           /* copy the tuple weights */
    if (tab_opc(table, TAB_COND|TAB_NORM) != 0)
      error(E_NOMEM);           /* compute one point coverages */
    r = tab_tplcnt(table);      /* get new number of data tuples */
    fprintf(stderr, "[%"TPLID_FMT, r);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    t = clock();                /* start timer, print message */
    fprintf(stderr, "summing possibility differences ... ");
    for (r = 0; r < n; r++) {   /* traverse the table again */
      tpl = tab_tpl(table, r);  /* copy tuple to the attribute set, */
      tpl_toas(tpl);            /* execute the graphical model and */
      gm_execx(gramod, rgm);    /* check one point coverages table */
      tab_possx(table, tpl, rtab);
      wgt = tpl_getxwgt(tpl);   /* get the (old) tuple weight and */
      res[0] += wgt *fabs(rgm[0] -rtab[0]);
      res[1] += wgt *fabs(rgm[1] -rtab[1]);
      res[2] += wgt *fabs(rgm[2] -rtab[2]);
      if (rgm[2] <= 0) imposs += wgt;
    }                           /* if the maximum possibility is 0, */
  }                             /* count the tuple as impossible */
  fprintf(stderr, "[%"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- compare to reference network --- */
  if (fn_ref) {                 /* if name of reference network given */

    /* --- parse domain descriptions --- */
    refas = as_create("domains", att_delete);
    if (!refas) error(E_NOMEM); /* create an attribute set */
    scan = scn_create();        /* create a scanner */
    if (!scan)  error(E_NOMEM); /* for the domain file */
    t = clock();                /* start timer, open input file */
    if (scn_open(scan, NULL, fn_ref) != 0)
      error(E_FOPEN, scn_name(scan));
    fprintf(stderr, "reading %s ... ", scn_name(scan));
    if (as_parse(refas, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan));
    as_setmark(attset, 0);      /* parse the domain descriptions */

    /* --- parse graphical model --- */
    refgm = gm_parse(attset, scan, type|GM_ALL);
    if (!refgm || !scn_eof(scan, 1))  /* parse the distributions */
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    scn_delete(scan, 1);        /* delete the scanner and */
    scan = NULL;                /* clear the scanner variable */

    /* --- compare graphical models --- */
    addcnt = miscnt = revcnt = 0;
    for (c = 0; c < m; c++) {   /* traverse the attributes */
      for (x = gm_concnt(refgm, c); --x >= 0; ) {
        z = gm_conid(refgm, c, x); /* traverse the conditions */
        for (y = gm_concnt(gramod, c); --y >= 0; )
          if (gm_conid(gramod, c, y) == z)
            break;              /* search for the current condition */
        if (y >= 0) continue;   /* if condition found, go to next */
        for (y = gm_concnt(gramod, z); --y >= 0; )
          if (gm_conid(gramod, z, y) == c)
            break;              /* search for the reversed condition */
        if (y >= 0) revcnt++;   /* if the condition was found, */
        else        miscnt++;   /* the edge is reversed, */
      }                         /* otherwise it is missing */
      for (x = gm_concnt(gramod, c); --x >= 0; ) {
        z = gm_conid(gramod, c, x); /* traverse the conditions */
        for (y = gm_concnt(refgm, c); --y >= 0; )
          if (gm_conid(refgm, c, y) == z)
            break;              /* search for the current condition */
        if (y >= 0) continue;   /* if condition found, go to next */
        for (y = gm_concnt(refgm, z); --y >= 0; )
          if (gm_conid(refgm, z, y) == c)
            break;              /* search for the reversed condition */
        if (y < 0) addcnt++;    /* if the condition was not found, */
      }                         /* there is an additional edge */
    }                           /* (sum differing conditions) */
  }

  /* --- print the results --- */
  t = clock();                  /* start timer, open output file */
  if (fn_out && (strcmp(fn_out, "-") == 0)) fn_out = "";
  if (fn_out && *fn_out) { out = fopen(fn_out, "w"); }
  else                   { out = stdout; fn_out = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_out);
  if (!out) error(E_FOPEN, fn_out);
  fprintf(out, "evaluation of %s on %s:\n", fn_net, fn_tab);
  fprintf(out, "number of attributes : %"ATTID_FMT"\n",
               gm_attcnt(gramod));
  fprintf(out, "number of conditions : %"ATTID_FMT"\n",
               gm_consum(gramod));
  fprintf(out, "number of parameters : %"ATTID_FMT"\n",
               gm_parsum(gramod));
  fprintf(out, "number of tuples     : %g\n", total);
  fprintf(out, "impossible tuples    : %g (%4.1f%%)\n", imposs,
         (total > 0) ? (imposs/total) *100 : 0);
  if (type == GM_PROB) {        /* print aggregated probabilities */
    fprintf(out, "log(average prob.)   : %g\n", res[0]);
    fprintf(out, "log(minimum prob.)   : %g\n", res[1]);
    fprintf(out, "log(maximum prob.)   : %g\n", res[2]); }
  else {                        /* print aggregated possibilities */
    fprintf(out, "average poss. sum    : %g\n", res[0]);
    fprintf(out, "minimum poss. sum    : %g\n", res[1]);
    fprintf(out, "maximum poss. sum    : %g\n", res[2]);
  }
  if (fn_ref) {                 /* if compared to reference network */
    fprintf(out, "comparison to %s:\n", fn_ref);
    fprintf(out, "additional edges: %"ATTID_FMT"\n", addcnt);
    fprintf(out, "missing    edges: %"ATTID_FMT"\n", miscnt);
    fprintf(out, "reversed   edges: %"ATTID_FMT"\n", revcnt);
  }                             /* print the edge statistics */
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_out);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
