/*----------------------------------------------------------------------
  File    : ptree1.c
  Contents: probability/possibility tree management
            base functions
  Author  : Christian Borgelt
  History : 1995.10.22 file created as ptree.c
            1995.10.29 function pt_conadd() added
            1995.11.02 function pt_conrem() added
            1996.03.15 number of levels restricted to PT_MAXCON+1
            1996.04.16 buffers for marginal distributions added
            1996.05.06 modified to sparse tree construction
            1996.11.27 type of counters changed to floating point
            1997.02.28 tree mode (PT_PROB/PT_POSS) added, redesign
            1997.03.05 automatic invalidation of aggregates added
            1998.02.13 links between nodes made possible (pt_link)
            1998.02.18 function pt_leafcnt() added
            1998.06.10 semantics of function pt_down changed
            1999.02.20 adapted to structure changes, assertions added
            1999.02.24 module based on attribute set management
            1999.04.25 function pt_concopy() added
            1999.04.30 some assertions added, pt_rand simplified
            1999.10.20 bug in function aggr() fixed (aggr(lvl+1, ..))
            2002.01.22 major redesign completed
            2002.02.24 function pt_getcntx() added, pt_exec() redesigned
            2002.03.11 number of tuples added for poss. normalization
            2002.04.10 parameter flags of function pt_parse removed
            2002.07.01 function pt_parcnt() redesigned
            2002.07.03 function pt_occur() added (was a #define)
            2002.07.16 parameter 'delnodes' added to function pt_clear
            2002.07.17 bug in function pt_link() fixed (lvl increment)
            2006.02.16 restriction of number of conditions removed
            2006.03.09 bug in function pt_create() fixed (level vector)
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "ptree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE   16            /* block size for level vector */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static PTNODE* getchild (PTLVL *lvl, PTNODE *parent, VALID index)
{                               /* --- get/create a child node */
  size_t s;                     /* size of created node */
  PTNODE *child;                /* retrieved/created child node */

  assert(lvl && (index >= 0));  /* check the function arguments */
  if (parent) child = parent->children[index];
  else        child = lvl->list;/* if a child already exists, */
  if (child) return child;      /* simply return it */
  s = (lvl->index < 0)          /* compute the size of the node */
    ? sizeof(PTLEAF) +(size_t)(lvl->cnt-1) *sizeof(WEIGHT)
    : sizeof(PTNODE) +(size_t)(lvl->cnt-1) *sizeof(PTNODE*);
  child = (PTNODE*)calloc(1,s); /* create a new node and */
  if (!child) return NULL;      /* initialize its fields */
  child->parent = parent;
  child->index  = index;
  child->succ   = lvl->list;    /* add the new node at the head */
  lvl->list     = child;        /* of the level list */
  if (parent)                   /* store the node in the parent */
    parent->children[index] = child;
  return child;                 /* return the created node */
}  /* getchild() */

/*--------------------------------------------------------------------*/

static int makepath (PTREE *pt)
{                               /* --- create path to current node */
  ATTID i;                      /* loop variable for attributes */
  VALID k;                      /* loop variable for values */
  PTLVL *lvl;                   /* to traverse the tree levels */

  assert(pt && !pt->curr);      /* check the function argument */
  for (lvl = pt->levels +(i = pt->pathlen); --i >= 0; )
    if ((--lvl)->node) break;   /* go up on the path to the node and */
  pt->curr = lvl->node;         /* find the deepest existing node */
  k = (pt->curr) ? (lvl++)->index : 0;
  for (i = pt->pathlen-i; --i >= 0; ) {
    pt->curr = getchild(lvl, pt->curr, k);
    if (!pt->curr) return -1;   /* go to the child node, */
    k = (lvl++)->index;         /* creating it if necessary, */
  }                             /* and note the next index */
  return 0;                     /* return 'ok' */
}  /* makepath() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

PTREE* pt_create (ATT *att, int type)
{                               /* --- create a prob./poss. tree */
  PTREE *pt;                    /* created prob./poss. tree */
  PTLVL *lvl;                   /* root level description */

  assert(att                    /* check the function argument */
  &&    (att_type(att) == AT_NOM));
  pt = (PTREE*)malloc(sizeof(PTREE));
  if (!pt) return NULL;         /* allocate memory and */
  pt->att     = att;            /* initialize fields */
  pt->type    = type & PT_POSS;
  pt->concnt  = pt->pathlen = 0;
  pt->fullcnt = pt->leafcnt = 1;
  pt->parcnt  = att_valcnt(att) -((pt->type & PT_POSS) ? 0 : 1);
  pt->curr    = NULL;
  pt->tplcnt  = -1;
  pt->occur   = -1;
  pt->total   = pt->lcorr = 0;
  pt->measure = PT_NONE;
  pt->eval    = 0;
  pt->buf     = NULL;
  pt->convsz  = BLKSIZE;        /* create the level vector */
  pt->levels  = lvl = (PTLVL*)malloc((BLKSIZE+1) *sizeof(PTLVL));
  if (!lvl) { free(pt); return NULL; }
  lvl->att    = att;            /* initialize the root level */
  lvl->cnt    = att_valcnt(att);
  lvl->index  = (type & PT_POSS)/* use the index for the leaf level */
              ? -2 : -1;        /* as an additional type indicator */
  lvl->list   = lvl->node = NULL;
  lvl->buf    = (WEIGHT*)malloc((size_t)lvl->cnt *2 *sizeof(WEIGHT));
  if (!lvl->buf) { free(lvl); free(pt); return NULL; }
  return pt;                    /* return the created tree */
}  /* pt_create() */

/*--------------------------------------------------------------------*/

void pt_delete (PTREE *pt, int delatts)
{                               /* --- delete prob./poss. tree */
  ATTID  i;                     /* loop variable for attributes */
  PTNODE *n, *t;                /* to traverse the nodes */

  assert(pt);                   /* check the function argument */
  for (i = pt->concnt+1; --i >= 0; ) {
    free(pt->levels[i].buf);    /* delete the level buffer */
    for (n = pt->levels[i].list; n; ) {
      t = n; n = n->succ; free(t); }
    if (delatts) att_delete(pt->levels[i].att);
  }                             /* delete all nodes of the tree, */
  if (pt->buf) free(pt->buf);   /* the attributes, the buffer, */
  free(pt->levels);             /* the level vector, */
  free(pt);                     /* and the base structure */
}  /* pt_delete() */

/*--------------------------------------------------------------------*/

void pt_clear (PTREE *pt, int delnodes)
{                               /* --- clear prob./poss. tree */
  ATTID  i;                     /* loop variable for attributes */
  PTLVL  *lvl;                  /* to traverse the tree levels */
  PTNODE *n, *t;                /* to traverse the nodes */
  PTLEAF *leaf;                 /* to traverse the leaves */

  assert(pt);                   /* check the function argument */
  if (!pt->levels->list) return;/* if the tree is empty, abort */
  if (pt->buf) {                /* delete an evaluation buffer */
    free(pt->buf); pt->buf = NULL; }
  pt->total = 0;                /* clear the counter total */
  if (delnodes) {               /* if to delete the nodes */
    pt->curr    = NULL;         /* the current node and */
    pt->pathlen =  0;           /* the path to it are invalid */
    pt->occur   = -1;           /* occurrence flags are valid */
    pt->parcnt  =  0;           /* number of parameters is invalid */
    for (i = pt->concnt +1; --i >= 0; ) {
      for (n = pt->levels[i].list; n; ) {
        t = n; n = n->succ; free(t); }
      pt->levels[i].list = NULL;/* delete all nodes of the tree and */
    } }                         /* clear the level list pointers */
  else {                        /* if only to clear the counters */
    pt->occur = 0;              /* occurrence flags are invalid */
    lvl = pt->levels +pt->concnt;
    for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
      leaf->total = 0;          /* traverse the leaf nodes */
      memset(leaf->cnts, 0, (size_t)lvl->cnt *sizeof(WEIGHT));
    }                           /* clear all leaf counters */
  }                             /* (but keep the nodes) */
}  /* pt_clear() */

/*--------------------------------------------------------------------*/

int pt_conadd (PTREE *pt, ATTID lvlid, ATT *att)
{                               /* --- add a condition to a tree */
  ATTID  k;                     /* loop variable, condition counter */
  VALID  n;                     /* number of attribute values */
  PTLVL  *lvl;                  /* to traverse the tree levels */
  WEIGHT *buf;                  /* buffer for new level */

  assert(pt                     /* check the function arguments */
  &&    (lvlid >= 0) && (lvlid <= pt->concnt)
  &&    att && (att_type(att) == AT_NOM));
  k = pt->convsz;               /* get the size of the level vector */
  if (pt->concnt >= k) {        /* if the level vector is full */
    k += (k > BLKSIZE) ? k >> 1 : BLKSIZE;
    lvl = (PTLVL*)realloc(pt->levels, (size_t)(k+1) *sizeof(PTLVL));
    if (!lvl) return -1;        /* enlarge the level vector */
    pt->levels = lvl; pt->convsz = k;
  }                             /* set new level vector and its size */
  n   = att_valcnt(att);        /* get the number of values/children */
  buf = (WEIGHT*)malloc((size_t)n *sizeof(WEIGHT));
  if (!buf) return -1;          /* allocate a buffer for the level */
  pt_clear(pt, 1);              /* delete all nodes of the tree */
  for (lvl = pt->levels +(k = ++pt->concnt); --k >= lvlid; ) {
    --lvl; lvl[1] = lvl[0]; }   /* shift down the tree levels */
  lvl->cnt   = n;               /* initialize the description */
  lvl->att   = att;             /* of the inserted level */
  lvl->list  = NULL;
  lvl->index = 0;
  lvl->buf   = buf;
  return 0;                     /* return 'ok' */
}  /* pt_conadd() */

/*--------------------------------------------------------------------*/

void pt_conrem (PTREE *pt, ATTID lvlid, int delatt)
{                               /* --- remove a condition from a tree */
  ATTID i;                      /* loop variable for attributes */
  PTLVL *lvl;                   /* to traverse the tree levels */

  assert(pt && (lvlid < pt->concnt));    /* check the arguments */
  pt_clear(pt, 1);              /* delete all nodes of the tree */
  if (lvlid < 0) {              /* if to remove all conditions */
    for (lvl = pt->levels +(i = pt->concnt); --i >= 0; ) {
      free((--lvl)->buf);       /* delete the level buffer */
      if (delatt) att_delete(lvl->att);
    }                           /* delete the condition attribute */
    lvl[0] = lvl[pt->concnt];   /* move the leaf level to the root */
    pt->concnt = 0; }           /* there are no conditions anymore */
  else {                        /* if to remove only one level */
    lvl = pt->levels +lvlid;    /* get the level to remove and */
    free(lvl->buf);             /* delete buffer and attribute */
    if (delatt) att_delete(lvl->att);
    for (i = pt->concnt -lvlid; --i >= 0; lvl++)
      lvl[0] = lvl[1];          /* shift up the deeper tree levels */
    pt->concnt--;               /* there is now one condition less */
  }
}  /* pt_conrem() */

/*--------------------------------------------------------------------*/

int pt_concopy (PTREE *dst, PTREE *src)
{                               /* --- copy conditions of a tree */
  ATTID i;                      /* loop variable */

  assert(dst && src);           /* check the function arguments */
  for (i = 0; i < src->concnt; i++)
    if (pt_conadd(dst, i, src->levels[i].att) != 0)
      return -1;                /* add all conditions from source */
  return 0;                     /* return 'ok' */
}  /* pt_concopy() */

/*--------------------------------------------------------------------*/

void pt_down (PTREE *pt, VALID index)
{                               /* --- go down in prob./poss. tree */
  PTLVL *lvl;                   /* the current tree level */

  assert(pt && !pt_atleaf(pt)   /* check the function arguments */
  &&    (index >= 0) && (index <= pt->levels[pt->pathlen].cnt));
  lvl = pt->levels +pt->pathlen++; /* get the current level and */
  lvl->index = index;              /* note the child index used */
  if (pt->curr)                 /* get the corresponding child node */
    pt->curr = pt->curr->children[index];
  (++lvl)->node = pt->curr;     /* note the node on the next level */
}  /* pt_down() */

/*--------------------------------------------------------------------*/

void pt_up (PTREE *pt, int mode)
{                               /* --- go up in prob./poss. tree */
  PTLVL  *lvl;                  /* the current tree level */
  PTNODE *node;                 /* to traverse the ancestor nodes */

  assert(pt);                   /* check the function argument */
  lvl = pt->levels;             /* get the level descriptions */
  if ((mode == PT_ROOT)         /* if to go to the root node */
  ||  (pt->pathlen <= 0)) {     /* or if there is no path, */
    pt->curr = lvl->list;       /* make the root node the */
    pt->pathlen = 0; return;    /* current node and clear the path */
  }
  lvl += --pt->pathlen;         /* get the parent level */
  if ((mode != PT_PARENT)       /* if not to go to the true parent */
  ||  !pt->curr)                /* or if there is no current node, */
    pt->curr = lvl->node;       /* go to the node from the path */
  else {                        /* if to go to the true parent, */
    pt->curr = node = pt->curr->parent;  /* get the true parent */
    while (node != lvl->node) { /* if not descended from the parent, */
      lvl->node      = node;    /* build a new path with true parent */
      (--lvl)->index = node->index & ~PT_OCCUR;
      node = node->parent;      /* and the index in the parent, */
    }                           /* then go up in the tree */
  }                             /* (loop must terminate at root) */
}  /* pt_up() */

/*--------------------------------------------------------------------*/

int pt_setcnt (PTREE *pt, VALID index, WEIGHT cnt)
{                               /* --- set a counter in current node */
  PTLEAF *leaf;                 /* leaf to modify */
  WEIGHT old;                   /* buffer for the old value */

  assert(pt && pt_atleaf(pt)    /* check the function arguments */
  &&    (index >= 0) && (index < pt->levels[pt->pathlen].cnt)
  &&    (cnt   >= 0));
  if (!pt->curr && (makepath(pt) != 0))
    return -1;                  /* create missing nodes if necessary */
  leaf = (PTLEAF*)pt->curr;     /* get the leaf to modify */
  if (pt->type & PT_POSS) {     /* if possibility tree, */
    leaf->cnts[index] = cnt;    /* set the new value and */
    pt->total = -1; }           /* invalidate the total */
  else {                        /* if probability tree, */
    old = leaf->cnts[index];    /* note the old counter value */
    leaf->cnts[index] = cnt;    /* and set the new value */
    leaf->total += cnt -= old;  /* adapt the totals of the leaf */
    pt->total   += cnt;         /* and of the whole tree */
  }
  return pt->occur = 0;         /* clear occ. flag and return 'ok' */
}  /* pt_setcnt() */

/*--------------------------------------------------------------------*/

WEIGHT pt_getcntx (PTREE *pt)
{                               /* --- get counter of current inst. */
  VALID  i;                     /* buffer for a value */
  ATTID  k;                     /* loop variable for attributes */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTNODE *node;                 /* to traverse the nodes */

  assert(pt);                   /* check the function arguments */
  lvl  = pt->levels;            /* get the level vector */
  node = lvl->list;             /* and the root node and */
  if (!node) return 0;          /* traverse the conditions */
  for (k = pt->concnt; --k >= 0; lvl++) {
    i = att_inst(lvl->att)->n;  /* get the instance of the condition */
    assert(i < lvl->cnt);       /* and check the attribute value */
    if (i < 0) return -1;       /* if the value is unknown, abort */
    node = node->children[i];   /* get the child corresponding to */
    if (!node) return 0;        /* the attribute value and */
  }                             /* if it is missing, abort */
  i = att_inst(lvl->att)->n;    /* get the attribute instance and */
  assert(i < lvl->cnt);         /* return the corresponding counter */
  return (i >= 0) ? ((PTLEAF*)node)->cnts[i] : -1;
}  /* pt_getcntx() */

/*--------------------------------------------------------------------*/

double pt_gettot (PTREE *pt)
{                               /* --- get aggregate of counters */
  assert(pt && pt_atleaf(pt));  /* check the function argument */
  if (!pt->curr)                /* if there is no current node, */
    return 0;                   /* all counters are zero (default) */
  if (pt->total < 0)            /* if the totals are invalid, */
    pt_total(pt);               /* compute the counter totals */
  return ((PTLEAF*)pt->curr)->total;
}  /* pt_gettot() */            /* return the leaf total */

/*--------------------------------------------------------------------*/

int pt_occur (PTREE *pt)
{                               /* --- check for nonzero counters */
  VALID  i;                     /* loop variable for values */
  ATTID  k;                     /* loop variable for attributes */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTNODE *node;                 /* to traverse the inner nodes */
  PTNODE **p;                   /* to traverse the child vectors */
  PTLEAF *leaf;                 /* to traverse the leaves */

  assert(pt);                   /* check the function argument */
  if (!pt->occur) {             /* if the flags are not set */
    if (pt->total < 0)          /* if the leaf totals are not valid, */
      pt_total(pt);             /* compute them first */
    lvl = pt->levels+pt->concnt;/* traverse the leaf nodes */
    for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
      if (leaf->total <= 0) leaf->index &= ~PT_OCCUR;
      else                  leaf->index |=  PT_OCCUR;
    }                           /* clear/set the node occurrence flag */
    for (k = pt->concnt; --k >= 0; ) {
      for (node = (--lvl)->list; node; node = node->succ) {
        for (p = node->children +(i = lvl->cnt); --i >= 0; )
          if (*--p && ((*p)->index & PT_OCCUR)) break;
        if (i < 0) node->index &= ~PT_OCCUR;
        else       node->index |=  PT_OCCUR;
      }                         /* traverse the other tree levels */
    }                           /* and set the occurrence flags */
    pt->occur = -1;             /* set the global flag */
  }
  return (pt->curr && (pt->curr->index & PT_OCCUR)) ? 1 : 0;
}  /* pt_occur() */             /* return the occurrence flag */

/*--------------------------------------------------------------------*/

ATTID pt_path (PTREE *pt, VALID *path)
{                               /* --- get path to current node */
  ATTID  i;                     /* loop variable */
  PTNODE *node;                 /* to traverse the parent nodes */
  PTLVL  *lvl;                  /* to traverse the path to the node */

  assert(pt && path);           /* check the function arguments */
  node = pt->curr;              /* start at the current node, */
  lvl  = pt->levels+pt->pathlen;/* which is the last on the path */
  for (path += (i = pt->pathlen); --i >= 0; ) {
    if (!node) {                /* if there is no current node, */
      *--path = (--lvl)->index; /* get the child index and the */
      node    = lvl->node; }    /* parent node from the path */
    else {                      /* if there is a current node, */
      *--path = node->index & ~PT_OCCUR;
      node    = node->parent;   /* note index in parent node */
    }                           /* and go to the parent node */
  }
  return pt->pathlen;           /* return the length of the path */
}  /* pt_path() */

/*--------------------------------------------------------------------*/

int pt_link (PTREE *pt, VALID index, const VALID *path)
{                               /* --- set a link to another node */
  ATTID  i;                     /* loop variable for conditions */
  VALID  k;                     /* buffer for a value/child index */
  PTLVL  *lvl;                  /* to traverse the level descriptions */
  PTNODE *node;                 /* to traverse the nodes on the path */

  assert(pt && !pt_atleaf(pt)   /* check the function arguments */
  &&    (index >= 0) && (index <= pt->levels[pt->pathlen].cnt));
  if (!pt->curr && (makepath(pt) != 0))
    return -1;                  /* create missing nodes if necessary */
  lvl  = pt->levels;            /* start at the root node and */
  node = lvl->list;             /* traverse the path to link to */
  for (i = pt->pathlen +1; --i >= 0; ) {
    k = (path) ? *path++ : lvl->link;
    assert(k < lvl->cnt);       /* get and check the child index */
    node = getchild(++lvl, node, k);
    if (!node) return -1;       /* go to the child node and */
  }                             /* create it if necessary */
  pt->curr->children[index] = node;
  pt->parcnt = 0;               /* install the link */
  return 0;                     /* return 'ok' */
}  /* pt_link() */

/*--------------------------------------------------------------------*/

VALID pt_parcnt (PTREE *pt)
{                               /* --- get the number of parameters */
  ATTID  i;                     /* loop variables for attributes */
  VALID  m, n;                  /* loop variable, number of leaves */
  PTLVL  *lvl;                  /* to traverse the level descriptions */
  PTNODE *node, **p;            /* to traverse the nodes/children */

  assert(pt);                   /* check the function argument */
  if (pt->parcnt > 0)           /* if the number of parameters */
    return pt->parcnt;          /* is valid, simply return it */
  lvl = pt->levels +pt->concnt; /* traverse the tree levels */
  for (n = 1, i = pt->concnt; --i >= 0; )
    n *= (--lvl)->cnt;          /* compute full number of leaves */
  pt->fullcnt = n; n = 0;       /* note the full number of leaves */
  if (lvl->list) {              /* if the tree is not empty, */
    lvl->list->pathcnt = 1;     /* start with one path at the root */
    for (i = pt->concnt; --i >= 0; lvl++) { /* traverse the levels */
      for (node = (lvl+1)->list; node; node = node->succ)
        node->pathcnt = 0;      /* clear all path counters */
      for (node = lvl->list; node; node = node->succ)
        for (p = node->children +(m = lvl->cnt); --m >= 0; )
          if (*--p) (*p)->pathcnt += node->pathcnt;
    }                           /* sum the number of paths to a node */
    for (node = lvl->list; node; node = node->succ)
      n += node->pathcnt -1;    /* count the number of leaves */
  }                             /* that are merged to other leaves */
  pt->leafcnt = pt->fullcnt -n; /* compute the number of leaves and */
  lvl = pt->levels +pt->concnt; /* get the number of params. per leaf */
  n   = lvl->cnt -((pt->type & PT_POSS) ? 0 : 1);
  return pt->parcnt = pt->leafcnt *n;
}  /* pt_parcnt() */            /* return the number of parameters */

/*--------------------------------------------------------------------*/

double pt_total (PTREE *pt)
{                               /* --- compute totals of counters */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the counters */
  double s = 0, t;              /* buffers for totals */

  assert(pt);                   /* check the function argument */
  if (pt->total >= 0)           /* if the total is valid, */
    return pt->total;           /* simply return it */
  lvl = pt->levels +pt->concnt; /* traverse the leaf nodes */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    c = leaf->cnts +(i = lvl->cnt);
    t = 0;                      /* traverse the counters */
    if (pt->type & PT_POSS) {   /* if possibility tree */
      while (--i >= 0) if (*--c > t) t = *c;
      if (t > s) s = t; }       /* find leaf maximum and global max. */
    else {                      /* if probability tree */
      while (--i >= 0) t += *--c;
      s += t;                   /* sum the counters in the leaf */
    }                           /* and all counters in the tree */
    leaf->total = t;            /* store the leaf total */
  }
  return pt->total = s;         /* return the computed total */
}  /* pt_total() */

/*--------------------------------------------------------------------*/

static int aggr (PTLVL *lvl, PTNODE *node, WEIGHT wgt, int type)
{                               /* --- recursively aggregate weight */
  VALID  i;                     /* loop variable, buffer for instance */
  PTNODE *child;                /* to traverse the children */
  WEIGHT *c;                    /* to traverse the counters */

  assert(lvl && node);          /* check the function arguments */
  i = att_inst(lvl->att)->n;    /* get the attribute instance */
  assert(i < lvl->cnt);         /* and check its value */
  if (lvl->index < 0) {         /* --- if the node is a leaf node */
    if (i >= 0) {               /* if the attribute value is known, */
      c = ((PTLEAF*)node)->cnts +i;  /* update the corresp. counter */
      if (type == PT_PROB) *c += wgt;
      else if (wgt > *c)   *c  = wgt; }
    else {                      /* if the attribute value is unknown */
      c = ((PTLEAF*)node)->cnts +(i = lvl->cnt);
      if (type & PT_POSS) {     /* if possibility tree */
        while (--i >= 0) if (wgt > *--c) *c = wgt; }
      else {                    /* if probability tree */
        ((PTLEAF*)node)->total += wgt; wgt /= (WEIGHT)lvl->cnt;
        while (--i >= 0) *c += wgt;
      }                         /* update all counters of the leaf */
    } }                         /* and the leaf total */
  else {                        /* --- if the node is an inner node */
    if (i >= 0) {               /* if the attribute value is known */
      node = getchild(++lvl, node, i);
      if (!node) return -1;     /* get the corresponding child node */
      aggr(lvl, node, wgt, type); }    /* and aggregate recursively */
    else {                      /* if the attribute value is unknown */
      if (type == PT_PROB)      /* for probability trees distribute */
        wgt /= (WEIGHT)lvl->cnt;/* the weight equally on all values */
      for (i = (lvl++)->cnt; --i >= 0; ) {
        child = getchild(lvl, node, i);
        if (!child) return -1;  /* traverse the child nodes */
        aggr(lvl, child, wgt, type);
      }                         /* aggregate recursively */
    }                           /* for all children */
  }
  return 0;                     /* return 'ok' */
}  /* aggr() */

/*--------------------------------------------------------------------*/

int pt_aggr (PTREE *pt, WEIGHT wgt, int distuv)
{                               /* --- aggregate weight of an inst. */
  int    r;                     /* result of function aggr() */
  VALID  i;                     /* buffer for an instance */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTNODE *node;                 /* to traverse the nodes */

  assert(pt);                   /* check the function argument */
  lvl  = pt->levels;            /* get the level vector */
  node = getchild(lvl, NULL, 0);/* get the root node or */
  if (!node) return -1;         /* create it if necessary */
  if (distuv                    /* if to distribute for unknowns */
  || (pt->type & PT_POSS)) {    /* or this is a possibility tree */
    r = aggr(lvl, node, wgt, pt->type);
    if (r == 0) {               /* aggregate recursively */
      if (pt->type & PT_POSS) pt->total = -1;
      else pt->total += wgt;    /* on success invalidate or */
      pt->occur = 0;            /* update the counter total */
    }                           /* and clear the occurrence flag */
    return r;                   /* return a success indicator */
  }
  while (1) {                   /* traverse the tree levels */
    i = att_inst(lvl->att)->n;  /* get the value of the condition */
    assert(i < lvl->cnt);       /* check for a valid value identifier */
    if (i < 0) return 1;        /* check for an unknown value */
    if (lvl->index < 0) break;  /* abort loop on the leaf level */
    node = getchild(++lvl, node, i);
    if (!node) return -1;       /* go to the child node or */
  }                             /* create it if necessary */
  ((PTLEAF*)node)->cnts[i] += wgt;
  ((PTLEAF*)node)->total   += wgt;
  pt->total += wgt;             /* aggregate the value */
  return pt->occur = 0;         /* return 'ok' */
}  /* pt_aggr() */

/*--------------------------------------------------------------------*/

VALID pt_rand (PTREE *pt, RANDFN randfn)
{                               /* --- compute a random value index */
  ATTID  k;                     /* loop variable for attributes */
  VALID  i, n, m;               /* loop variables for values */
  PTLVL  *lvl;                  /* to traverse the tree levels */
  PTNODE *node;                 /* to traverse the prob./poss. tree */
  WEIGHT *c;                    /* to traverse the counters */
  WEIGHT *buf;                  /* to access the counter buffer */
  double sum;                   /* sum of counters/random value */

  assert(pt                     /* check the function arguments */
  &&    (pt->type == PT_PROB) && randfn);
  lvl  = pt->levels;            /* get the level vector */
  node = lvl->list;             /* and the root node and */
  if (!node) return 0;          /* traverse the conditions */
  for (k = pt->concnt; --k >= 0; lvl++) {
    i = att_inst(lvl->att)->n;  /* get the value of the condition */
    assert(i < lvl->cnt);       /* and check its validity */
    if (i < 0) return -1;       /* check for an unknown value */
    node = node->children[i];   /* get the corresponding child node */
    if (!node) return  0;       /* if it does not exist, abort */
  }
  sum = 0; c = ((PTLEAF*)node)->cnts;
  buf = lvl->buf; n = lvl->cnt; /* traverse the counter vector and */
  for (i = 0; i < n; i++)       /* build a distribution function */
    buf[i] = (WEIGHT)(sum += c[i] +pt->lcorr);
  sum *= randfn();              /* generate a random value and */
  for (i = 0; i < n; ) {        /* do a binary search in the */
    m = (i+n) >> 1;             /* distribution function to find */
    if (sum >= buf[m]) i = m+1; /* the corresponding value index */
    else               n = m;   /* after this search it is */
  }                             /* buf[i-1] <= sum < buf[i], */
  return i;                     /* so return the index i */
}  /* pt_rand() */

/*--------------------------------------------------------------------*/

double pt_exec (PTREE *pt)
{                               /* --- execute a prob./poss. tree */
  VALID  i;                     /* buffer for an instance */
  ATTID  k;                     /* loop variable for attributes */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTNODE *node;                 /* to traverse the nodes */
  double t;                     /* temporary buffer */

  assert(pt);                   /* check the function arguments */
  if (pt->total < 0)            /* if the totals are not valid, */
    pt_total(pt);               /* compute them (for normalization) */
  lvl  = pt->levels;            /* get the tree levels and */
  node = lvl->list;             /* traverse the conditions */
  for (k = pt->concnt; node && (--k >= 0); lvl++) {
    i = att_inst(lvl->att)->n;  /* get the instance of the condition */
    assert(i < lvl->cnt);       /* and check the attribute value */
    if (i < 0) return -1;       /* if the value is unknown, abort */
    node = node->children[i];   /* get the child corresponding */
  }                             /* to the attribute value */
  if (!node) {                  /* if no leaf could be reached, */
    if (pt->type & PT_POSS)     /* return a default value */
      t = (double)((pt->tplcnt >= 0) ? pt->tplcnt : pt->total)
        + (double)pt->lcorr;
    else t = (double)pt->levels[pt->concnt].cnt *(double)pt->lcorr;
    return (double)pt->lcorr / (t +DBL_MIN);
  }
  i = att_inst(lvl->att)->n;    /* get the attribute instance */
  assert(i < lvl->cnt);         /* and check the attribute value */
  if (i < 0) return -1;         /* if the value is unknown, abort */
  if (pt->type & PT_POSS)       /* get the normalization value */
    t = (double)((pt->tplcnt >= 0) ? pt->tplcnt : pt->total)
      + (double)pt->lcorr;
  else t = ((PTLEAF*)node)->total +(double)lvl->cnt *pt->lcorr;
  return (((PTLEAF*)node)->cnts[i] +pt->lcorr) / (t +DBL_MIN);
}  /* pt_exec() */              /* return rel. freq./deg. of poss. */

/*--------------------------------------------------------------------*/

static WEIGHT poss (PTLVL *lvl, PTNODE *node)
{                               /* --- recursively det. poss. degree */
  VALID  i;                     /* loop variable, buffer for instance */
  PTNODE **p;                   /* to traverse the vector of children */
  WEIGHT *c;                    /* to traverse the counters */
  WEIGHT max, t;                /* degree of possibility, buffer */

  assert(lvl);                  /* check the function argument */
  if (!node) return 0;          /* if there is no node, abort */
  i = att_inst(lvl->att)->n;    /* get the attribute instance */
  assert(i < lvl->cnt);         /* and check the value */
  if (lvl->index < 0) {         /* --- if the node is a leaf */
    if (i >= 0)                 /* if the attribute value is known, */
      return ((PTLEAF*)node)->cnts[i];  /* return the corr. counter */
    max = 0;                    /* otherwise traverse the counters */
    for (c = ((PTLEAF*)node)->cnts +(i = lvl->cnt); --i >= 0; ) {
      if (*--c > max) max = *c; /* determine the maximum */
    } }                         /* of all counters in the leaf */
  else {                        /* --- if the node is an inner node */
    if (i >= 0)                 /* if the attribute value is known, */
      return poss(lvl+1, node->children[i]);     /* go to the child */
    max = 0;                    /* otherwise traverse the children */
    for (p = node->children +(i = (lvl++)->cnt); --i >= 0; ) {
      t = poss(lvl, *--p);      /* recursively process each child */
      if (t > max) max = t;     /* determine the maximum over */
    }                           /* all children of the node */
  }
  return max;                   /* return the computed maximum */
}  /* poss() */

/*--------------------------------------------------------------------*/

double pt_poss (PTREE *pt)
{                               /* --- det. degree of possibility */
  assert(pt);                   /* check the function argument */
  return ((double)poss(pt->levels, pt->levels[0].list)
          + (double)pt->lcorr)
       / (((pt->tplcnt >= 0) ? pt->tplcnt : pt->total)
          + (double)pt->lcorr +DBL_MIN);
}  /* pt_poss() */              /* determine maximum recursively */
